import argparse
import os
import re
import shutil


class FullAction(argparse.Action):

    def __init__(self, option_strings, dest, default=None, required=False, help=None, metavar=None):
        super(FullAction, self).__init__(option_strings=option_strings, dest=dest, nargs=0, const=True,
                                         default=default, required=required, help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        if len(namespace.spider) > 0:
            setattr(namespace, self.dest, self.const)
        else:
            parser.error('must used with spider')


class CreateCommand:
    parser = None

    def add_arguments(self):
        parser = argparse.ArgumentParser(description="resolve")
        # subparsers = parser.add_subparsers(help='sub-command help')
        # parser_create = subparsers.add_parser('create', help='add help')

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-s", "--spider", nargs=1, help="创建爬虫 如 hoopa create -s <spider_name>"
        )
        group.add_argument(
            "-p", "--project", nargs=1, help="创建项目 如 hoopa create -p <project_name>"
        )

        parser.add_argument(
            "-f", "--full", help="full spider 如 hoopa create -s <spider_name> -f", default=False, action=FullAction
        )

        self.parser = parser

    def create_spider(self, template_path, spider_name, project=None):
        spider_class_name = spider_name
        if spider_class_name.islower():
            spider_class_name = spider_class_name.title().replace("_", "")

        with open(template_path, "r", encoding="utf-8") as file:
            spider_template = file.read()
        spider_template = spider_template.replace("${spider_name}", spider_name)
        spider_template = spider_template.replace("${spider_class_name}", spider_class_name)
        result = self._save_spider_to_file(spider_template, spider_name, project)

        return result

    @staticmethod
    def copy_callback(src, dst, *, follow_symlinks=True):
        if src.endswith(".py"):
            with open(src, "r", encoding="utf-8") as src_file, open(
                dst, "w", encoding="utf8"
            ) as dst_file:
                content = src_file.read()
                dst_file.write(content)
        else:
            shutil.copy2(src, dst, follow_symlinks=follow_symlinks)

    def create_project(self, template_path, project_name):
        if os.path.exists(project_name):
            print("%s 项目已经存在" % project_name)
            return

        shutil.copytree(template_path, project_name, copy_function=self.copy_callback)

        template_path = self._find_template("project_spider")
        getattr(self, f"create_spider")(template_path, project_name, project_name)

        return True

    def run_cmd(self):
        args = self.parser.parse_args()
        full = args.full

        if args.spider:
            spider_name = args.spider[0]
            create_type = "spider"
        elif args.project:
            spider_name = args.project[0]
            create_type = "project"
        else:
            raise
        # 检查spider_name
        if not re.search("^[a-zA-Z][a-zA-Z0-9_]*$", spider_name):
            raise Exception("命名不规范，请用下划线命名或驼峰命名方式")

        template_path = self._find_template(create_type, full)
        result = getattr(self, f"create_{create_type}")(template_path, spider_name)
        if result:
            print(f"{create_type} {spider_name} create success")

    @staticmethod
    def _find_template(template, full=False):
        if template == "spider":
            if full:
                template_file = "spider_full_template.tmpl"
            else:
                template_file = "spider_template.tmpl"
        else:
            if template == "project_spider":
                template_file = "project_spider_template.tmpl"
            else:
                template_file = "project_template"

        template_path = os.path.abspath(
            os.path.join(__file__, "../../templates", template_file)
        )
        if os.path.exists(template_path):
            return template_path

        print("Unable to find template: %s\n" % template)

    def _save_spider_to_file(self, spider, spider_name, project_name):
        spider_underline = self._cover_to_underline(spider_name)
        if project_name:
            spider_file = f"{project_name}/{spider_underline}.py"
        else:
            spider_file = f"{spider_underline}.py"

        if os.path.exists(spider_file):
            print("文件已存在")
            return

        with open(spider_file, "w", encoding="utf-8") as file:
            file.write(spider)

        return True

    @staticmethod
    def _cover_to_underline(key):
        regex = "[A-Z]*"
        capitals = re.findall(regex, key)

        if capitals:
            for pos, capital in enumerate(capitals):
                if not capital:
                    continue
                if pos == 0:
                    if len(capital) > 1:
                        key = key.replace(capital, capital.lower() + "_", 1)
                    else:
                        key = key.replace(capital, capital.lower(), 1)
                else:
                    if len(capital) > 1:
                        key = key.replace(capital, "_" + capital.lower() + "_", 1)
                    else:
                        key = key.replace(capital, "_" + capital.lower(), 1)

        return key



# encoding: utf-8
"""
爬虫核心
"""
import asyncio
import inspect
import operator
import time
import traceback
import typing
from copy import deepcopy
from signal import SIGINT, SIGTERM
from types import AsyncGeneratorType, CoroutineType

from loguru import logger

from hoopa.core.downloadermiddleware import DownloaderMiddleware
from hoopa.pipelines import PipelineManager
from hoopa.core.spidermiddleware import SpiderMiddleware
from hoopa.utils.concurrency import run_function, run_function_no_concurrency, iterate_in_threadpool
from hoopa.utils.log import Logging
from hoopa.utils.asynciter import AsyncIter
from hoopa.utils.helpers import split_list, get_timestamp, load_object, create_instance_and_init
from hoopa.exceptions import InvalidCallbackResult, Error, InvalidCallback
from hoopa.item import Item
from hoopa.request import Request
from hoopa.response import Response
from hoopa.core.scheduler import Scheduler
from hoopa.utils.project import get_project_settings
from hoopa.utils import decorators

try:
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


class Engine:
    def __init__(self, spider, loop=None):
        self.loop = loop
        asyncio.set_event_loop(self.loop)

        self.spider = spider

        # 循环获取request，默认True
        self.spider.run = True

        self.request_queue: asyncio.Queue = asyncio.Queue()

        # 重试中的个数
        self.retrying = 0

        # 请求数统计
        self.requests_count = 0

        # 读取配置文件，如果config下面没有配置文件，那么返回默认配置
        setting = get_project_settings(spider.settings_path)
        # 加载爬虫里面的配置，覆盖配置文件的配置
        setting.init_settings(spider)

        self.setting = setting
        self.spider.setting = setting
        Logging(setting)

    async def _load(self):
        # 初始化 spider init
        await run_function(self.spider.open_spider)
        await run_function(self.spider.init)

        # 初始化 download
        downloader_cls = load_object(self.setting["DOWNLOADER_CLS"])
        self.downloader = await create_instance_and_init(downloader_cls, self)

        # 初始化 scheduler
        self.scheduler = await create_instance_and_init(Scheduler, self)

        # 初始化stats
        self.stats = self.scheduler.stats

        # 初始化下载中间件
        self.downloader_middleware = await DownloaderMiddleware.create(self)
        # 初始化爬虫中间件
        self.spider_middleware = await SpiderMiddleware.create(self)
        # 初始化管道
        self.pipeline_manager = await PipelineManager.create(self)

        # 打印配置日志
        self.setting.print_log(self)

        # failure_to_waiting
        if self.spider.failure_to_waiting:
            await self.scheduler.failure_to_waiting(self.spider)

        # 处理start_requests
        await self._handle_start_requests()

    async def _process_async_callback(self, request: Request, response: Response, callback_results: AsyncGeneratorType):
        if not callback_results or response.ok != 1:
            return

        request_list = []

        async for callback_result in callback_results:
            if isinstance(callback_result, Request):
                request_list.append(callback_result)
            elif isinstance(callback_result, Item):
                await self.pipeline_manager.process_pipelines(request, response, callback_result, self.spider)
            elif isinstance(callback_result, list) and all(isinstance(item, Item) for item in callback_result):
                # 如果是item list，也进行处理
                await self.pipeline_manager.process_pipelines(request, response, callback_result, self.spider)
            else:
                callback_result_name = type(callback_result).__name__
                raise InvalidCallbackResult(f"<Parse invalid callback result type: {callback_result_name}>")

        # 处理新请求, 默认50个请求批量存储
        request_list_split = split_list(request_list, 50)
        for item_requests in request_list_split:
            count = await self.scheduler.add(item_requests)
            logger.debug(f"{request} push request {count}")

    async def _process_callback(self, request, response):
        # 如果response.ok != 1，请求失败，不进行回调
        if response.ok != 1:
            return

        # 响应码在处理列表内，处理响应
        try:
            process_func = getattr(self.spider, request.callback, None)
        except:
            raise InvalidCallback("解析函数错误")

        try:
            if process_func is not None:
                callback_results = await self.spider_middleware.scrape_response(process_func, request, response,
                                                                                self.spider)
                # 判断是否是异步生成器，生成器需要需要在线程池里运行，避免阻塞
                if callback_results and not inspect.isasyncgen(callback_results):
                    return iterate_in_threadpool(callback_results)
                else:
                    return callback_results
            else:
                raise Exception(f"<Parse invalid callback result type: {request.callback}>")
        except Exception as e:
            response.ok = 0
            response.error = Error(e, traceback.format_exc())
            logger.error(f"{request} {response} callback error \n{response.error.stack}")

    @decorators.handle_download_callback_retry()
    async def handle_download_callback(self, request: Request):
        """
        处理请求
        @param request: request对象
        """
        # 请求数+1
        self.requests_count += 1

        # 加载request中间件, 并调用下载器
        response = await self.downloader_middleware.download(self.downloader.fetch, request, self.spider)

        # 回调
        callback_result = await self._process_callback(request, response)

        # 处理异步返回的request和item
        await self._process_async_callback(request, response, callback_result)

        return response

    @decorators.timeout_it()
    async def _process_task(self, request: Request):
        """
        处理请求
        @param request: request对象
        """
        response = Response()
        task_request = deepcopy(request)
        try:
            # 处理请求和回调
            response = await self.handle_download_callback(task_request)
            # 处理请求结果
            await self.scheduler.set_result(request, response, task_request)
        except Exception as e:
            response.ok = -1
            response.error = Error(e, traceback.format_exc())
            logger.error(f"{request} {response} callback error \n{response.error.stack}")

        if response.ok != 1:
            await run_function(self.spider.process_failed, task_request, response)
        else:
            await run_function(self.spider.process_succeed, task_request, response)

    async def _handle_start_requests(self):
        """
        用于初始化url，默认读取start_urls, 可重写
        """
        callback_results = await run_function(self.spider.start_requests)

        if isinstance(callback_results, typing.Generator):
            callback_results = AsyncIter(callback_results)

        request_list = []

        if callback_results and not isinstance(callback_results, CoroutineType):
            async for callback_result in callback_results:
                if isinstance(callback_result, Request):
                    request_list.append(callback_result)

        # 处理新请求, 默认50个请求批量存储
        request_list_split = split_list(request_list, self.spider.push_number)
        for item_requests in request_list_split:
            count = await self.scheduler.add(item_requests)
            logger.debug(f"start_requests push request {count}")

    async def start_worker(self):
        """
        Start spider worker
        :return:
        """
        while True:
            request_item = await self.request_queue.get()
            asyncio.run_coroutine_threadsafe(self._process_task(request_item), loop=self.loop)
            self.request_queue.task_done()

    async def consumer(self):
        logger.debug(f"consumer started")
        last_time = time.time()
        while self.spider.run:
            qsize = self.request_queue.qsize()
            add_task_count = max(self.spider.worker_numbers - qsize - self.retrying, 0)
            # logger.error(f"add_task_count: {add_task_count} {qsize} {self.retrying}")
            add_task_count = min(add_task_count, self.spider.worker_numbers)
            for _ in range(add_task_count):
                request_item = await self.scheduler.get(self.spider.priority)
                if request_item is None:
                    break
                self.request_queue.put_nowait(request_item)

            # 休眠
            sleep_second = max(self.spider.download_delay - (time.time() - last_time), 0.01)

            await asyncio.sleep(sleep_second)
            last_time = time.time()
            await self.scheduler.check_scheduler(self.spider)

    async def _start_spider(self):
        """
        启动爬虫，不断从队列中获取请求任务
        """
        # 初始化
        await self._load()

        if self.spider.run:
            logger.info(f"Spider start")

        workers = [
            asyncio.ensure_future(self.start_worker())
            for _ in range(self.spider.worker_numbers * 3)
        ]

        logger.debug(f"Worker started: {len(workers)}")

        await self.request_queue.join()

        await self.consumer()

        await self.finish()

    async def _start(self, before_start=None, after_stop=None):
        # 添加信号
        # SIGINT：由Interrupt Key产生，通常是CTRL+C或者DELETE。发送给所有ForeGround Group的进程
        # SIGTERM：请求中止进程，kill命令缺省发送
        for signal in (SIGINT, SIGTERM):
            try:
                self.loop.add_signal_handler(signal, lambda: asyncio.ensure_future(self.cancel_all_tasks(signal)))
            except NotImplementedError:
                pass

        # 爬虫开始之前执行
        await self.spider.run_spider_hook(before_start)

        # 爬虫开始
        try:
            await self._start_spider()
        except KeyboardInterrupt:
            await self.cancel_all_tasks()
        finally:
            # 爬虫结束后执行
            await self.spider.run_spider_hook(after_stop)
            logger.info("Spider finished!")

    @classmethod
    async def async_start(cls, before_start=None, after_stop=None, loop=None):
        loop = loop or asyncio.get_event_loop()
        spider_ins = cls(loop=loop)
        await spider_ins._start(before_start, after_stop)
        return spider_ins

    @classmethod
    def start(cls, before_start=None, after_stop=None, loop=None):
        """
        爬虫入口, 非异步
        @rtype: object
        """
        loop = loop or asyncio.new_event_loop()
        spider_ins = cls(loop=loop)
        loop.run_until_complete(spider_ins._start(before_start, after_stop))
        spider_ins.loop.run_until_complete(spider_ins.loop.shutdown_asyncgens())
        spider_ins.loop.close()
        return spider_ins

    async def finish(self):
        start_time = await self.stats.get_value("start_time")
        finish_time = int(get_timestamp())
        await self.stats.max_value("finish_time", finish_time)
        await self.stats.set_value("total_time", finish_time - start_time)

        spider_stats = await self.stats.get_stats()
        spider_stats_sorted_keys = sorted(spider_stats.items(), key=operator.itemgetter(0))
        blank = " " * 4
        body = "\nstats\n"

        body += f"\n".join(f"{blank}{k:50s}: {v}" for k, v in spider_stats_sorted_keys)
        logger.info(body)

        await run_function(self.spider.close_spider, spider_stats)

        await self.close()

    async def close(self):
        await self.scheduler.close()
        await run_function_no_concurrency(self.downloader.close)
        await run_function_no_concurrency(self.downloader_middleware.close)
        await run_function_no_concurrency(self.spider_middleware.close)
        await run_function_no_concurrency(self.pipeline_manager.close)
        await self.cancel_all_tasks()

    async def cancel_all_tasks(self, _signal=None):
        logger.info(f"Stopping spider: {self.spider.name}")
        tasks = []
        for task in asyncio.all_tasks():
            if task is not asyncio.tasks.current_task():
                tasks.append(task)
                task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)



import pickle
from copy import deepcopy

import ujson


def dumps(data, module=None):
    if not module:
        module = ujson
    return module.dumps(data)


def loads(data, module=None):
    if not module:
        module = ujson
    return module.loads(data)


def serialize_request_and_response(request, response):
    request_kwargs = request.replace_to_kwargs

    for item in deepcopy(request_kwargs):
        if not isinstance(item, (int, bool, str, dict, list)):
            request_kwargs.pop(item)

    response_dict = response.serialize()
    response_dict.setdefault("request", request_kwargs)

    return pickle.dumps(response_dict)



import asyncio
import hoopa


async def downloader_demo():
    downloader = hoopa.AiohttpDownloader()
    response = await downloader.fetch(hoopa.Request(url="https://httpbin.org/get"))
    print(response)


asyncio.run(downloader_demo())



# encoding: utf-8
"""
爬虫核心
"""
import weakref
from abc import ABC
from inspect import isawaitable
import typing
import asyncio

import ujson
from loguru import logger

from hoopa.core.engine import Engine
from hoopa.settings import const, Setting
from hoopa.utils.connection import get_aio_redis
from hoopa.utils.helpers import get_md5
from hoopa.exceptions import SpiderHookError
from hoopa.request import Request
from hoopa.response import Response
from hoopa.item import Item


try:
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


class BaseSpider:
    """
    used for extend spider

    Public attributes:
    - name: 爬虫名称.
    - worker_numbers: worker_numbers任务数，每download_delay秒worker_numbers个， 默认每1秒1个
    - download_delay: 爬虫请求间隔，每download_delay秒worker_numbers个， 默认每1秒1个
    - pending_threshold: pending超时时间.
    - run_forever: 任务完成不停止, 默认False.
    - queue_cls: 任务队列路径，默认：const.MemoryQueue(hoopa.queues.MemoryQueue).
    - clean_queue: 清空任务队列，默认False.
    - priority: 指定队列优先级，redis优先队列有效.
    - downloader_cls: 下载器路径，默认：const.AiohttpDownloader(hoopa.downloader.AiohttpDownloader).
    - downloader_middlewares: 下载中间件
    - spider_middlewares: 爬虫中间件
    - pipelines: 管道
    - dupefilter_cls: 去重器路径，默认MemoryDupeFilter，另外有RedisDupeFilter
    - clean_dupefilter: 清空去重器，默认等于clean_queue
    - dupefilter_setting: 去重器设置，默认等于redis_setting
    - redis_setting: redis连接配置，可以是字典，也可以是uri 例如："redis://127.0.0.1:6379/0?encoding=utf-8"
    - serialization: 序列化模块，默认ujson，可选pickle
    - log_config： 自定义logger.configure的参数，类型为字典
    - log_level： 日志级别，默认INFO
    - log_write_file： 日志是否写入文件，默认否
    - settings_path： 默认配置文件位置
    - start_urls： 起始url列表
    - interrupt_with_error： 出现错误时推出，默认False
    - failure_to_waiting：  将错误队列放入等待队列，默认False
    - push_number：  请求推送到redis单次最大数量
    - run:  控制爬虫停止，默认为True运行，设置为False停止
    """
    name: str = "hoopa"
    worker_numbers: int = None
    download_delay: int = None
    pending_threshold: int = None
    run_forever: bool = None
    queue_cls: str = None
    clean_queue: bool = None
    priority: int = None
    downloader_cls: str = None
    http_client_kwargs: bool = None
    downloader_middlewares: list = None
    spider_middlewares: list = None
    pipelines: list = None
    dupefilter_cls: bool = None
    clean_dupefilter: bool = None
    stats_cls: str = None
    dupefilter_setting: typing.Union[dict, str] = None
    redis_setting: typing.Union[dict, str] = None
    log_config: dict = None
    log_level: str = None
    log_write_file: bool = None
    serialization: bool = None
    settings_path: str = "config.settings"
    start_urls: list = []
    interrupt_with_error: bool = None
    setting: Setting = None
    push_number: int = None
    failure_to_waiting: bool = None
    run: bool = None

    async def run_spider_hook(self, hook_func):
        """
        Run hook before/after spider start crawling
        :param hook_func: aws function
        :return:
        """
        if callable(hook_func):
            try:
                aws_hook_func = hook_func(weakref.proxy(self))
                if isawaitable(aws_hook_func):
                    await aws_hook_func
            except Exception as e:
                raise SpiderHookError(f"<Hook {hook_func.__name__}: {e}")

    async def open_spider(self):
        pass

    async def close_spider(self, spider_stats):
        pass

    async def init(self):
        """
        由于魔术方法__init__无法使用await，可以在爬虫重写这个方法，以达到初始化的目的
        """
        pass

    async def start_requests(self):
        """
        用于初始化url，默认读取start_urls, 可重写
        """
        for url in self.start_urls:
            yield Request(url=url, callback=self.parse)

    async def process_item(self, item: Item):
        return item

    async def process_request(self, request: Request):
        return request

    async def parse(self, request: Request, response: Response):
        raise NotImplementedError("<!!! parse function is expected !!!>")

    async def process_failed(self, request: Request, response: Response):
        """
        处理错误的任务
        :param request: Request
        :param response: Response
        :return:
        """
        pass

    async def process_succeed(self, request: Request, response: Response):
        """
        处理成功的任务
        :param request: Request
        :param response: Response
        :return:
        """
        pass

    @classmethod
    async def async_start(cls, before_start=None, after_stop=None, loop=None):
        loop = loop or asyncio.get_event_loop()
        spider_ins = cls()
        engine = Engine(spider_ins, loop=loop)
        await engine._start(before_start, after_stop)
        return engine

    @classmethod
    def start(cls, before_start=None, after_stop=None, loop=None):
        """
        爬虫入口, 非异步
        @rtype: object
        """
        loop = loop or asyncio.new_event_loop()
        spider_ins = cls()
        engine = Engine(spider_ins, loop=loop)
        loop.run_until_complete(engine._start(before_start, after_stop))
        engine.loop.run_until_complete(engine.loop.shutdown_asyncgens())
        engine.loop.close()
        return engine


class Spider(BaseSpider, ABC):

    async def process_failed(self, request: Request, response: Response):
        """
        处理错误的任务
        :param request: Request
        :param response: Response
        :return:
        """
        pass

    async def process_succeed(self, request: Request, response: Response):
        """
        处理成功的任务
        :param request: Request
        :param response: Response
        :return:
        """
        pass

    async def process_item(self, item: typing.Union[Item, list[Item]]):
        """
        处理item，爬虫里面的process_item，可以直接调用
        @param item:
        """
        return item

    async def process_request(self, request):
        """
        爬虫里面的载中间件，处理下载前的request
        @param request:
        """
        pass


class RedisSpider(BaseSpider, ABC):
    redis_pool = None
    queue_cls = const.RedisQueue

    async def open_spider(self):
        self.redis_pool = await get_aio_redis(self.setting["REDIS_SETTING"])
        self.setting.set("DUPEFILTER_CLS", const.RedisDupeFilter, "default")
        self.setting.set("STATS_CLS", const.RedisStatsCollector, "default")
        self.setting.set("DUPEFILTER_SETTING", self.setting.get("REDIS_SETTING"), "default")
        self.setting.set("CLEAN_DUPEFILTER", self.setting.get("CLEAN_QUEUE"), "default")

    async def process_item(self, item: Item):
        """
        处理item，RedisQueue的时候默认保存到redis。其他情况需要重写此方法
        @rtype: bool
        @param item:
        """
        item_str_json = ujson.dumps(item.values)
        item_md5 = get_md5(item_str_json)

        with await self.redis_pool as conn:
            result = await conn.hmset(f"{self.name}:{item.item_name}", item_md5, item_str_json)
        if result:
            logger.info(f"{item.values}")





# encoding: utf-8
import importlib.util
import os

from hoopa.settings import Setting


def get_project_settings(settings_path="config.settings"):
    """
    读取配置文件，先从默认配置读取，然后读取自定义配置
    @param settings_path: 配置文件可以填写相对路径，例如config.settings。也可以填写绝对路径：D:/setting.py
    @return: Setting对象
    """

    setting = Setting()
    setting.set_module("hoopa.settings.default_settings", priority="default")
    try:
        if os.path.isfile(settings_path):
            module_spec = importlib.util.spec_from_file_location(os.path.basename(settings_path), settings_path)
            module = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module)
            setting.set_module(module, priority="project")
        else:
            setting.set_module(settings_path)
    except Exception as e:
        setting.project_setting = False

    return setting



import sys
import cProfile

from hoopa.exceptions import UsageError
from hoopa.commands.create import CreateCommand


def _pop_command_name(argv):
    i = 0
    for arg in argv[1:]:
        if not arg.startswith('-'):
            del argv[i]
            return arg
        i += 1


def _print_unknown_command(cmd_name):
    print("Unknown command: %s\n" % cmd_name)
    print('Use "hoopa" to see available commands')


def _run_print_help(parser, func, *a, **kw):
    try:
        func(*a, **kw)
    except UsageError as e:
        if str(e):
            parser.error(str(e))
        if e.print_help:
            parser.print_help()
        sys.exit(2)


def _run_command(cmd, args, opts):
    if opts.profile:
        _run_command_profiled(cmd, args, opts)
    else:
        cmd.run(args, opts)


def _run_command_profiled(cmd, args, opts):
    if opts.profile:
        sys.stderr.write("scrapy: writing cProfile stats to %r\n" % opts.profile)
    loc = locals()
    p = cProfile.Profile()
    p.runctx('cmd.run(args, opts)', globals(), loc)
    if opts.profile:
        p.dump_stats(opts.profile)


def _print_commands():
    # with open(join(dirname(dirname(__file__)), "VERSION"), "rb") as f:
    #     version = f.read().decode("ascii").strip()
    #
    # print("hoopa {}".format(version))
    print("Usage:")
    print("  hoopa <command> [options] [args]\n")
    print("Available commands:")
    cmd_list = {"create": "create project、spider、item and so on"}
    for cmd_name, cmd_class in sorted(cmd_list.items()):
        print("  %-13s %s" % (cmd_name, cmd_class))

    print('Use "hoopa <command> -h" to see more info about a command')


def execute(argv=None):
    if argv is None:
        argv = sys.argv

    if len(argv) < 2:
        _print_commands()
        return

    cmd_name = argv.pop(1)
    cmd_list = {
        "create": CreateCommand
    }

    if not cmd_name:
        _print_commands()
        sys.exit(0)
    elif cmd_name not in cmd_list:
        _print_unknown_command(cmd_name)
        sys.exit(2)

    cmd = cmd_list[cmd_name]()
    cmd.add_arguments()
    cmd.run_cmd()

    sys.exit()


if __name__ == '__main__':
    execute()




# encoding: utf-8
"""
request对象
"""

from types import SimpleNamespace
from typing import (Any, Iterable, Mapping, Optional, Union, Callable)

from aiohttp import ClientTimeout
from aiohttp.helpers import (BasicAuth, sentinel)
from aiohttp.client_reqrep import Fingerprint
from aiohttp.typedefs import StrOrURL, LooseCookies, LooseHeaders

from httpx._types import (AuthTypes, CookieTypes, HeaderTypes, QueryParamTypes, RequestContent, RequestData,
                          RequestFiles, TimeoutTypes, URLTypes,)
from httpx._config import (UNSET, UnsetType)
from loguru import logger

from w3lib.url import add_or_replace_parameters, canonicalize_url

try:
    from ssl import SSLContext
except ImportError:  # pragma: no cover
    SSLContext = object  # type: ignore

from hoopa.utils import helpers
from hoopa.utils.serialization import loads, dumps


not_serialize_params = ["session", "message", 'http_kwargs', "retries"]

serialize_params = ['url', 'headers', 'method', 'params', 'data', 'json', 'meta', 'dont_filter', 'priority',
                    'callback', 'client_kwargs', 'http_kwargs']

not_kwargs_list = ["session", "message", "callback", "dont_filter", "meta", "priority",
                   "client_kwargs", "http_kwargs", "retries", "retry_times", "retry_delay"]


class AiohttpParams:
    method: str = None
    url: StrOrURL = None
    params: Optional[Mapping[str, str]] = None
    data: Any = None
    json: Any = None
    cookies: Optional[LooseCookies] = None
    headers: Optional[LooseHeaders] = None
    skip_auto_headers: Optional[Iterable[str]] = None
    auth: Optional[BasicAuth] = None
    allow_redirects: bool = True
    max_redirects: int = 10
    compress: Optional[str] = None
    chunked: Optional[bool] = None
    expect100: bool = False
    raise_for_status: Optional[bool] = None
    read_until_eof: bool = True
    proxy: Optional[StrOrURL] = None
    proxy_auth: Optional[BasicAuth] = None
    timeout: Union[ClientTimeout, object] = sentinel
    verify_ssl: Optional[bool] = None
    fingerprint: Optional[bytes] = None
    ssl_context: Optional[SSLContext] = None
    ssl: Optional[Union[SSLContext, bool, Fingerprint]] = None
    proxy_headers: Optional[LooseHeaders] = None
    trace_request_ctx: Optional[SimpleNamespace] = None
    read_bufsize: Optional[int] = None


class HttpxParams:
    method: str = None
    url: URLTypes = None
    content: RequestContent = None
    data: RequestData = None
    files: RequestFiles = None
    json: Any = None
    params: QueryParamTypes = None
    headers: HeaderTypes = None
    cookies: CookieTypes = None
    auth: Union[AuthTypes, UnsetType] = UNSET
    allow_redirects: bool = True
    timeout: Union[TimeoutTypes, UnsetType] = UNSET


class RequestParams:
    method = None
    url = None
    params = None
    data = None
    headers = None
    cookies = None
    files = None
    auth = None
    timeout = None
    allow_redirects = True
    proxies = None
    hooks = None
    stream = None
    verify = None
    cert = None
    json = None


class OtherParams:
    # rabbitmq的message
    message = None
    # session参数， http请求的session
    session = None
    # 重试次数
    retry_times = 3
    # 重试时的间隔
    retry_delay = 0

    http_kwargs = None


class Request(AiohttpParams, HttpxParams, RequestParams, OtherParams):
    def __init__(
            self,
            url,
            method='get',
            headers=None,
            params=None,
            data=None,
            json=None,
            cookies=None,
            timeout=10,
            allow_redirects=True,
            callback=None,
            meta=None,
            dont_filter=False,
            priority=0,
            retry_times=3,
            retry_delay=1,
            client_kwargs=None,
            **_http_kwargs
    ):
        self.url = url
        self.headers = headers
        self.method = method
        self.params = params
        self.data = data
        self.json = json
        self.allow_redirects = allow_redirects
        self.cookies = cookies
        self.timeout = timeout

        # 其他参数
        for key, value in _http_kwargs.items():
            self.set(key, value)

        self.meta = meta

        # 默认为False，去重
        self.dont_filter = dont_filter

        # 优先级，越大优先级越大
        self.priority = priority

        # callback支持两种方式传入，可是函数名，也可以是函数，最终存储的是函数名字符串
        self.callback = callback if not callback or isinstance(callback, str) else callback.__name__

        self.client_kwargs = client_kwargs if client_kwargs else {}

        self.retry_times = retry_times
        self.retry_delay = retry_delay
        self.retries = 0

    def __getitem__(self, item):
        return getattr(self, item)

    def set(self, name, value):
        setattr(self, name, value)

    @staticmethod
    def unserialize(data_str, module=None):
        data = loads(data_str, module)
        http_kwargs = data.pop("http_kwargs")
        _request = Request(**data)
        for k, v in http_kwargs.items():
            _request.set(k, v)
        return _request

    def serialize(self, module=None):
        request_dict = {}
        _http_kwargs = {}
        for name, value in vars(self).items():
            if name not in not_serialize_params:
                # 不在request参数里面的参数都设置到http_kwargs
                if name in serialize_params:
                    request_dict.setdefault(name, value)
                else:
                    _http_kwargs[name] = value
        request_dict["http_kwargs"] = _http_kwargs
        return dumps(request_dict, module)

    def copy(self):
        kwargs = vars(self)
        cls = kwargs.pop('cls', self.__class__)
        return cls(**kwargs)

    @property
    def replace_to_kwargs(self):
        """
        生成请求参数
        @return:
        """
        _kwargs = {}
        for name, value in vars(self).items():
            if value is not None and name not in not_kwargs_list:
                # 不在not_kwargs_list里面的参数都设置到http_kwargs
                _kwargs.setdefault(name, value)

        # 把params字典的值统一转为str类型
        _params = _kwargs.get("params", None)
        if _params:
            _new_params = {}
            for k, v in _params.items():
                _new_params[k] = str(v)
            _kwargs["params"] = _new_params
        return _kwargs

    @property
    def fp(self):
        return helpers.request_fingerprint(self)

    @property
    def request_url(self):
        _url = self.url
        if self.params:
            _url = add_or_replace_parameters(self.url, self.params)

        return canonicalize_url(_url)

    def __repr__(self):
        return f"<Request p{self.priority} {self.method} {self.request_url}>"



import asyncio
import functools
import typing
from asyncio import iscoroutinefunction
from typing import Any, AsyncGenerator, Iterator, Dict

try:
    import contextvars  # Python 3.7+ only.
except ImportError:  # pragma: no cover
    contextvars = None  # type: ignore

T = typing.TypeVar("T")


async def run_in_threadpool(
    func: typing.Callable[..., T], *args: typing.Any, **kwargs: typing.Any
) -> T:
    loop = asyncio.get_event_loop()
    if contextvars is not None:  # pragma: no cover
        # Ensure we run in the same context
        child = functools.partial(func, *args, **kwargs)
        context = contextvars.copy_context()
        func = context.run
        args = (child,)
    elif kwargs:  # pragma: no cover
        # loop.run_in_executor doesn't accept 'kwargs', so bind them in here
        func = functools.partial(func, **kwargs)
    return await loop.run_in_executor(None, func, *args)


async def run_function(callable_fun,  *args, **kwargs) -> Any:
    if iscoroutinefunction(callable_fun):
        return await callable_fun(*args, **kwargs)
    else:
        return await run_in_threadpool(callable_fun, *args, **kwargs)


async def run_function_no_concurrency(func,  *args, **kwargs):
    if iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)


class _StopIteration(Exception):
    pass


def _next(iterator: Iterator) -> Any:
    # We can't raise `StopIteration` from within the threadpool iterator
    # and catch it outside that context, so we coerce them into a different
    # exception type.
    try:
        return next(iterator)
    except StopIteration:
        raise _StopIteration


async def iterate_in_threadpool(iterator: Iterator) -> AsyncGenerator:
    while True:
        try:
            yield await run_in_threadpool(_next, iterator)
        except _StopIteration:
            break


from loguru import logger


class HandleProcessError:
    async def process_exception(self, request, response, item, error, spider_ins):
        logger.error(f"{item} 入库失败")
        response.ok = 0
        response.error = error
        logger.error(error.stack)





#!/usr/bin/env python
import traceback

from hoopa.exceptions import Error, InvalidOutput
from hoopa.middleware import MiddlewareManager
from hoopa.request import Request
from hoopa.response import Response
from hoopa.utils.asynciter import AsyncIter
from hoopa.utils.concurrency import run_function


class DownloaderMiddleware(MiddlewareManager):
    """下载中间件"""

    @classmethod
    def _get_mw_list_from_engine(cls, engine):
        return engine.setting.get("DOWNLOADER_MIDDLEWARES") + engine.setting.get("DOWNLOADER_MIDDLEWARES_BASE")

    def _add_middleware(self, mw):
        super()._add_middleware(mw)
        if hasattr(mw, "process_request") and callable(getattr(mw, "process_request")):
            self.methods["process_request"].append(mw.process_request)
            self.names["process_request"].append(mw.__class__.__name__)

        if hasattr(mw, "process_response") and callable(getattr(mw, "process_response")):
            self.methods["process_response"].append(mw.process_response)
            self.names["process_response"].append(mw.__class__.__name__)

        if hasattr(mw, "process_exception") and callable(getattr(mw, "process_exception")):
            self.methods["process_exception"].append(mw.process_exception)
            self.names["process_exception"].append(mw.__class__.__name__)

    async def process_request(self, request: Request, spider_ins):
        methods = [spider_ins.process_request] + list(self.methods["process_request"])
        for method in methods:
            # 主要是为了处理spider里面的process_item参数的不同
            if method.__code__.co_argcount == 2:
                response = await run_function(method, request)
            else:
                response = await run_function(method, request, spider_ins)

            if response is not None and not isinstance(response, (Response, Request)):
                raise InvalidOutput(f"<Middleware {method.__name__}: must return None, Response or Request")
            if response:
                return response

    async def process_response(self, request: Request, response: Response, spider_ins):
        # 如果response是request，返回异步生成器
        if isinstance(response, Request):
            return AsyncIter([response])

        for method in self.methods["process_response"]:
            middleware_response = await run_function(method, request, response, spider_ins)
            if middleware_response is not None and not isinstance(middleware_response, (Response, Request)):
                raise InvalidOutput(f"<Middleware {method.__name__}: must return None, Response or Request")
            if isinstance(middleware_response, Request):
                return AsyncIter([middleware_response])
            if isinstance(middleware_response, Response):
                response = middleware_response
                break

        return response

    async def process_exception(self, request, exception, spider_ins):
        for method in self.methods['process_exception']:
            middleware_response = await run_function(method, request, exception, spider_ins)
            if middleware_response is not None and not isinstance(middleware_response, (Response, Request)):
                raise InvalidOutput(
                    f"Middleware {method.__self__.__class__.__name__}"
                    ".process_exception must return None, Response or "
                    f"Request, got {type(middleware_response)}"
                )
            if middleware_response:
                return middleware_response
        raise exception

    async def download(self, download_func, request, spider_ins):
        # 加载request中间件, 并调用下载器
        response = await self.process_request(request, spider_ins)

        # 没有response，调用下载器
        if not response:
            # 执行完request_middleware，调用下载器
            try:
                response = await run_function(download_func, request)
            except Exception as e:
                return await self.process_exception(request, Error(e, traceback.format_exc()), spider_ins)

        if response:
            # 加载response中间件
            process_response_result = await self.process_response(request, response, spider_ins)

            if process_response_result:
                return process_response_result

        return response



#!/usr/bin/env python
import traceback
from copy import deepcopy

from hoopa.item import Item
from hoopa.exceptions import Error, InvalidOutput
from hoopa.middleware import MiddlewareManager
from hoopa.utils.concurrency import run_function


class PipelineManager(MiddlewareManager):
    """下载中间件"""

    @classmethod
    def _get_mw_list_from_engine(cls, engine):
        return engine.setting.get("PIPELINES") + engine.setting.get("PIPELINES_BASE")

    def _add_middleware(self, mw):
        super()._add_middleware(mw)

        if hasattr(mw, "process_item") and callable(getattr(mw, "process_item")):
            self.methods["process_item"].append(mw.process_item)
            self.names["process_item"].append(mw.__class__.__name__)

        if hasattr(mw, "process_exception") and callable(getattr(mw, "process_exception")):
            self.methods["process_exception"].append(mw.process_exception)
            self.names["process_exception"].append(mw.__class__.__name__)

    async def process_item(self, item, spider_ins):
        _item = deepcopy(item)

        methods = [spider_ins.process_item] + list(self.methods["process_item"])

        for method in methods:
            # 主要是为了处理spider里面的process_item参数的不同
            if method.__code__.co_argcount == 2:
                _item = await run_function(method, _item)
            else:
                _item = await run_function(method, _item, spider_ins)

            if not _item:
                break

            if not isinstance(_item, Item) and (not (isinstance(_item, list) and all(isinstance(item, Item) for item in _item))):
                raise InvalidOutput(
                    f"Middleware {method.__self__.__class__.__name__}"
                    ".process_item must return None or "
                    f"Item or list[Item], got {type(_item)}"
                )

    async def process_exception(self, request, response, item, exception, spider_ins):
        for method in self.methods['process_exception']:
            middleware_response = await run_function(method, request, response, item, exception, spider_ins)
            if middleware_response is not None and not isinstance(middleware_response, bool):
                raise InvalidOutput(
                    f"Middleware {method.__self__.__class__.__name__}"
                    ".process_exception must return None or "
                    f"bool, got {type(middleware_response)}"
                )
            if middleware_response:
                return

    async def process_pipelines(self, request, response, item, spider_ins):
        # 执行完request_middleware，调用下载器
        try:
            await self.process_item(item, spider_ins)
        except InvalidOutput:
            raise
        except Exception as e:
            return await self.process_exception(request, response, item, Error(e, traceback.format_exc()), spider_ins)



"""
Item，数据对象需要继承这个
"""
from copy import deepcopy
from typing import Dict


class Item:
    def __init__(self, item_name=None, data: Dict = None):
        """
        @param item_name: item名称
        @param data: 字典，可以直接把字典设置到item
        """

        self._item_name = item_name if item_name else type(self).__name__  # 使用类名作为默认值

        if data:
            self.__dict__.update(**data)

    @property
    def values(self):
        data = deepcopy(self.__dict__)
        data.pop("_item_name")
        return data

    @property
    def item_name(self):
        return self._item_name

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __repr__(self):
        return f"<{self.item_name} {self.values}>"



# encoding: utf-8
"""
去重器
"""

from hoopa.utils.connection import get_aio_redis


class BaseDupeFilter:
    async def get(self, fp):
        pass

    async def add(self, fp):
        pass

    async def clean_queue(self):
        pass

    async def close(self):
        pass


class MemoryDupeFilter(BaseDupeFilter):
    """
    基于内存去重
    """
    def __init__(self, *args, **kwargs):
        self.pool = set()

    async def get(self, fp):
        result = fp in self.pool
        return result is False

    async def add(self, fp):
        self.pool.add(fp)


class RedisDupeFilter(BaseDupeFilter):
    """
    redis去重，pika，tendis等兼容redis的数据库
    """

    def __init__(self, dupefilter_setting, key, engine):
        self.pool = None
        self.dupefilter_setting = dupefilter_setting
        self.key = key
        self.engine = engine

    @classmethod
    async def create(cls, engine):
        dupefilter_setting = engine.setting["DUPEFILTER_SETTING"]
        key = f"{engine.setting['NAME']}:DupeFilter"
        return cls(dupefilter_setting, key, engine)

    async def init(self):
        self.pool = await get_aio_redis(self.dupefilter_setting)

    async def get(self, fp):
        with await self.pool as conn:
            is_member = await conn.sismember(self.key, fp)
            return is_member == 0

    async def add(self, fp):
        with await self.pool as conn:
            added = await conn.sadd(self.key, fp)
            return added == 0

    async def clean_queue(self):
        with await self.pool as conn:
            return await conn.delete(self.key)

    async def close(self):
        self.pool.close()



from loguru import logger

from hoopa import Response


class HandleHttpErrorMiddleware:
    async def process_response(self, request, response, spider_ins):
        # if response.status != 200:
        #     response.ok = 0
        #     return response
        pass

    async def process_exception(self, request, error, spider_ins):
        logger.error(f"{request} fetch error \n {error.stack}")
        return Response(error=error, ok=0)



# encoding: utf-8
"""
调度器
"""

import time
import typing

from loguru import logger
from w3lib.url import is_url

from hoopa.exceptions import InvalidUrl
from hoopa.request import Request
from hoopa.response import Response
from hoopa.utils.helpers import get_timestamp, load_object, create_instance_and_init


class Scheduler:
    """
    调度器
    """
    def __init__(self, dupefilter=None, scheduler_queue=None, stats=None, engine=None):
        self.scheduler_queue = scheduler_queue
        self.dupefilter = dupefilter
        self.stats = stats
        self.engine = engine

        self._last_check_status_time = time.time()

    @classmethod
    async def create(cls, engine):
        dupefilter_cls = load_object(engine.setting["DUPEFILTER_CLS"])
        dupefilter = await create_instance_and_init(dupefilter_cls, engine)

        scheduler_queue_cls = load_object(engine.setting["QUEUE_CLS"])
        scheduler_queue = await create_instance_and_init(scheduler_queue_cls, engine)

        stats_cls = load_object(engine.setting["STATS_CLS"])
        stats = await create_instance_and_init(stats_cls, engine)

        return cls(dupefilter, scheduler_queue, stats, engine)

    async def init(self):
        # 删除队列
        if self.engine.setting["CLEAN_QUEUE"]:
            # 获取相关的key
            await self.scheduler_queue.clean_queue()

        # 删除去重队列
        if self.engine.setting["CLEAN_DUPEFILTER"]:
            # 获取相关的key
            await self.dupefilter.clean_queue()

        # 初始化爬虫开始时间
        await self.stats.min_value("start_time", int(get_timestamp()))

    async def get(self, priority=None):
        """
        从队列中获取一个 request
        @param priority: 权重，取出对应权重的request，当队列为redis时生效
        """
        if priority and not isinstance(priority, (int, list)):
            raise TypeError(f"queue_priority must be int or list, not {type(priority)}")

        request = await self.scheduler_queue.get(priority)

        if request:
            await self.stats.inc_value('queue/request_count', 1)
            logger.debug(f"get request {request}")
        return request

    async def add(self, requests: typing.Union[Request, typing.List[Request]]):
        """
        向队列添加多个request
        @param requests:
        """
        if not isinstance(requests, list):
            requests = [requests]

        # 判断url是否合法
        for item in requests:
            if item is None:
                raise InvalidUrl(f"Invalid url: {item} url is None ")
            if not is_url(item.url):
                raise InvalidUrl(f"Invalid url: {item.url} ")

        request_stats = {}
        # 去重
        request_list = []
        for request in requests:
            if request.dont_filter:
                request_stats[request.priority] = request_stats.get(request.priority, 0) + 1
                request_list.append(request)
            else:
                if await self.dupefilter.get(request.fp):
                    request_stats[request.priority] = request_stats.get(request.priority, 0) + 1
                    request_list.append(request)
        #  去重后为空
        if not request_list:
            return 0

        # 放进队列
        set_len = await self.scheduler_queue.add(request_list)

        # 把放进队列的request去重
        for request in requests:
            if not request.dont_filter:
                await self.dupefilter.add(request.fp)

        # 统计，统计去重数（待定）
        # 统计request
        for k, v in request_stats.items():
            await self.stats.inc_value(f"request/priority_count/{k}", v)

        return set_len

    async def set_result(self, request: Request, response: Response, task_request: Request):
        """
        保存结果
        @param request:
        @param response:
        @param task_request: 爬虫请求过程中的request
        """
        await self.scheduler_queue.set_result(request, response, task_request)

        if response.ok == 1:
            logger.debug(f"{request} result: success, status: {response.status}")
        else:
            logger.error(f"{request} result: failure, status: {response.status}")

        # 统计信息
        await self.stats.inc_value(f"queue/response_count", 1)
        await self.stats.inc_value(f'queue/response_count/priority_{request.priority}/{response.ok}', 1)

    async def check_scheduler(self, spider_ins):
        await self.scheduler_queue.check_status(spider_ins, self.engine.spider.setting["RUN_FOREVER"])

    async def failure_to_waiting(self, spider_ins):
        await self.scheduler_queue.failure_to_waiting(spider_ins)

    async def close(self):
        await self.dupefilter.close()
        await self.scheduler_queue.close()
        await self.stats.close()



from loguru import logger

from hoopa.utils.url import get_location_from_history


class HandleHttpSuccessMiddleware:
    async def process_response(self, request, response, spider_ins):
        if response.ok == 1:
            if response.history:
                last_url = get_location_from_history(response.history)
                logger.debug(f"{request} redirect <{last_url}> success")
            else:
                logger.debug(f"{request} fetch {response}")







#!/usr/bin/env python
import traceback

from hoopa.item import Item
from hoopa.exceptions import Error, InvalidOutput
from hoopa.middleware import MiddlewareManager
from hoopa.request import Request
from hoopa.response import Response
from hoopa.utils.concurrency import run_function


class SpiderMiddleware(MiddlewareManager):
    """爬虫中间件"""

    @classmethod
    def _get_mw_list_from_engine(cls, engine):
        return engine.setting.get("SPIDER_MIDDLEWARES") + engine.setting.get("SPIDER_MIDDLEWARES_BASE")

    def _add_middleware(self, mw):
        super()._add_middleware(mw)
        if hasattr(mw, "process_request") and callable(getattr(mw, "process_request")):
            self.methods["process_request"].append(mw.process_request)
            self.names["process_request"].append(mw.__class__.__name__)

        if hasattr(mw, "process_response") and callable(getattr(mw, "process_response")):
            self.methods["process_response"].appendleft(mw.process_response)
            self.names["process_response"].appendleft(mw.__class__.__name__)

        if hasattr(mw, "process_exception") and callable(getattr(mw, "process_exception")):
            self.methods["process_exception"].appendleft(mw.process_exception)
            self.names["process_exception"].appendleft(mw.__class__.__name__)

    async def process_input(self, request: Request, response: Response, spider_ins):
        for method in self.methods["process_request"]:
            middleware_response = await run_function(method, request, response, spider_ins)
            if response is not None and not isinstance(middleware_response, bool):
                raise Exception(f"<Middleware {method.__name__}: must return None, got {type(response)}")

            if middleware_response:
                return

    async def process_output(self, request: Request, response: Response, result, spider_ins):
        for method in self.methods["process_response"]:
            middleware_response = await run_function(method, request, response, result, spider_ins)
            if middleware_response is not None and not isinstance(middleware_response, (Response, Request)):
                raise InvalidOutput(f"<Middleware {method.__name__}: must return None, Request or Item, "
                                    f"got {type(response)}")

            if isinstance(middleware_response, (Request, Item)):
                return middleware_response

        return result

    async def process_exception(self, request, response, error, spider_ins):
        for method in self.methods['process_exception']:
            middleware_response = await run_function(method, request, response, error, spider_ins)
            if middleware_response is not None and not isinstance(middleware_response, bool):
                raise InvalidOutput(
                    f"Middleware {method.__self__.__class__.__name__}"
                    ".process_exception must return None or "
                    f"bool, got {type(middleware_response)}"
                )
            if middleware_response:
                return
        raise error.exception

    async def scrape_response(self, parse_func, request, response, spider_ins):
        # 调用中间件
        await self.process_input(request, response, spider_ins)

        try:
            results = await run_function(parse_func, request, response)

            if results:
                # 加载response中间件
                process_response_results = []
                for result in results:
                    process_response_result = await self.process_output(request, response, result, spider_ins)
                    process_response_results.append(process_response_result)

                return iter(process_response_results)

        except Exception as e:
            await self.process_exception(request, response, Error(e, traceback.format_exc()), spider_ins)




# encoding: utf-8
from hoopa import const
from hoopa.downloadermiddlewares.handle_http_error import HandleHttpErrorMiddleware
from hoopa.downloadermiddlewares.handle_http_success import HandleHttpSuccessMiddleware
from hoopa.downloadermiddlewares.stats import StatsMiddleware
from hoopa.pipelinemiddlewares.default_pipeline import DefaultPipeline
from hoopa.pipelinemiddlewares.handle_process_error import HandleProcessError
from hoopa.spidermiddlewares.handle_parse_error import HandleParseErrorMiddleware

NAME = "hoopa"

# 最大协程数
WORKER_NUMBERS = 1
# 请求间隔, 可以是两个int组成的list，间隔随机取两个数之间的随机浮点数
DOWNLOAD_DELAY = 3
# pending超时时间，超过这个时间，放回waiting
PENDING_THRESHOLD = 100
# 任务完成不停止
RUN_FOREVER = False
INTERRUPT_WITH_ERROR = False
# 失败队列重新爬取
FAILURE_TO_WAITING = False
PUSH_NUMBER = 100


# 队列
# 调度器队列，默认redis, memory, mq
QUEUE_CLS = const.MemoryQueue
# 删除队列（包括数据集，去重队列）
CLEAN_QUEUE = False
# 指定优先级，仅当队列为redis有用
PRIORITY = None

# 下载器aiohttp httpx
DOWNLOADER_CLS = const.AiohttpDownloader
HTTP_CLIENT_KWARGS = None

# 下载中间件
# 执行顺序： DOWNLOADER_MIDDLEWARES
DOWNLOADER_MIDDLEWARES = [
]

DOWNLOADER_MIDDLEWARES_BASE = [
    StatsMiddleware,
    HandleHttpErrorMiddleware,
    HandleHttpSuccessMiddleware,
]


# 爬虫中间件
# 执行顺序： DOWNLOADER_MIDDLEWARES
SPIDER_MIDDLEWARES = [
]

SPIDER_MIDDLEWARES_BASE = [
    HandleParseErrorMiddleware
]

# pipelines
# 执行顺序：DOWNLOADER_MIDDLEWARES
PIPELINES = [
]

PIPELINES_BASE = [
    DefaultPipeline,
    HandleProcessError
]


# redis配置信息
# REDIS_SETTING = "redis://127.0.0.1:6379/0?encoding=utf-8"
REDIS_SETTING = {
    'host': '127.0.0.1',
    'port': 6379,
    'db': 0,
    'password': ''
}

# 默认去重，不删除去重队列, 将根据queue的类型来决定
DUPEFILTER_CLS = const.MemoryDupeFilter
# 是否删除去重队列
CLEAN_DUPEFILTER = None
# 去重数据库连接配置
DUPEFILTER_SETTING = None

# 统计器, 默认内存
STATS_CLS = const.MemoryStatsCollector

# 其他配置
# 序列化: pickle, ujson, orjson
SERIALIZATION = "ujson"

# 日志配置
LOG_CONFIG = None        # 自定义logger.configure的参数，类型为字典
LOG_LEVEL = "DEBUG"
LOG_WRITE_FILE = False



"""
collecting spider stats
"""

from hoopa.utils.connection import get_aio_redis


class StatsCollector:

    def __init__(self):
        self._stats = {}

    async def get_value(self, key, default=None, spider=None):
        return self._stats.get(key, default)

    async def get_stats(self, spider=None):
        return self._stats

    async def set_value(self, key, value, spider=None):
        self._stats[key] = value

    async def set_stats(self, stats, spider=None):
        self._stats = stats

    async def inc_value(self, key, count=1, start=0, spider=None):
        d = self._stats
        d[key] = d.setdefault(key, start) + count

    async def max_value(self, key, value, spider=None):
        self._stats[key] = max(self._stats.setdefault(key, value), value)

    async def min_value(self, key, value, spider=None):
        self._stats[key] = min(self._stats.setdefault(key, value), value)

    async def close(self):
        pass


class MemoryStatsCollector(StatsCollector):

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.spider_stats = {}


class DummyStatsCollector(StatsCollector):

    def get_value(self, key, default=None, spider=None):
        return default

    def set_value(self, key, value, spider=None):
        pass

    def set_stats(self, stats, spider=None):
        pass

    def inc_value(self, key, count=1, start=0, spider=None):
        pass

    def max_value(self, key, value, spider=None):
        pass

    def min_value(self, key, value, spider=None):
        pass


class RedisStatsCollector(StatsCollector):
    """
    Stats Collector based on Redis
    """
    pool = None
    stats_key = None

    def __init__(self, redis_setting, stats_key, engine):
        super().__init__()
        self.redis_setting = redis_setting
        self.stats_key = stats_key
        self.engine = engine

    @classmethod
    async def create(cls, engine):
        redis_setting = engine.setting["REDIS_SETTING"]
        stats_key = f"{engine.setting['NAME']}:Stats"

        return cls(redis_setting, stats_key, engine)

    async def init(self):
        self.pool = await get_aio_redis(self.redis_setting)

    async def get_value(self, key, default=None, spider=None):
        """Return the value of hash stats"""
        with await self.pool as conn:
            if await conn.hexists(self.stats_key, key):
                return int(await conn.hget(self.stats_key, key))
            else:
                return default

    async def get_stats(self, spider=None):
        """Return the all of the values of hash stats"""
        with await self.pool as conn:
            return await conn.hgetall(self.stats_key)

    async def set_value(self, key, value, spider=None):
        """Set the value according to hash key of stats"""
        with await self.pool as conn:
            return await conn.hset(self.stats_key, key, value)

    async def set_stats(self, stats, spider=None):
        """Set all the hash stats"""
        with await self.pool as conn:
            await conn.hmset(self.stats_key, stats)

    async def inc_value(self, key, count=1, start=0, spider=None):
        """Set increment of value according to key"""
        with await self.pool as conn:
            if not await conn.hexists(self.stats_key, key):
                await self.set_value(key, start)
            await conn.hincrby(self.stats_key, key, count)

    async def max_value(self, key, value, spider=None):
        """Set max value between current and new value"""
        await self.set_value(key, max(await self.get_value(key, value), value))

    async def min_value(self, key, value, spider=None):
        """Set min value between current and new value"""
        await self.set_value(key, min(await self.get_value(key, value), value))

    async def close(self):
        self.pool.close()



from importlib import import_module
from typing import Dict, Any

from loguru import logger

from hoopa.settings import const
from hoopa.settings.const import const_map

SETTINGS_PRIORITIES = {
    'default': 0,
    'project': 1,
    'spider': 2,
}

spider_attr_list = [
    'name',
    'worker_numbers',
    'download_delay',
    'pending_threshold',
    'run_forever',
    'queue_cls',
    'clean_queue',
    'priority',
    'downloader_cls',
    'http_client_kwargs',
    'dupefilter_cls',
    'clean_dupefilter',
    'dupefilter_setting',
    'redis_setting',
    'stats_cls',
    'downloader_middlewares',
    'spider_middlewares',
    'pipelines',
    'log_config',
    'log_level',
    'log_write_file',
    'serialization',
    'interrupt_with_error',
    'push_number',
    'failure_to_waiting'
]


class SettingItem:
    default: Any = None
    project: Any = None
    spider: Any = None

    @property
    def max_priority(self):
        if self.spider:
            return 2
        elif self.project:
            return 1
        else:
            return 0

    def get(self, priority):
        return getattr(self, priority)

    def set(self, priority, value):
        return setattr(self, priority, value)


class Setting:
    priority_map: Dict[str, SettingItem] = {}
    project_setting: bool = True

    # __getitem__和__setitem__，让对象可以像字典一样调用
    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def get(self, name, default=None, priority=None):
        if not priority:
            return getattr(self, name, default)

        return getattr(self.priority_map[name], priority, default)

    def set(self, name, value, priority="project"):
        setting_item = self.priority_map.get(name, SettingItem())
        if getattr(setting_item, priority, None) is not None:
            if SETTINGS_PRIORITIES[priority] >= self.priority_map[name].max_priority:
                setattr(self, name, value)
                setattr(self.priority_map[name], priority, value)
        else:
            setattr(self, name, value)
            setting_item.set(priority, value)
            self.priority_map[name] = setting_item

    def set_module(self, module, priority="project"):
        if isinstance(module, str):
            module = import_module(module)
        for key in dir(module):
            if key.isupper():
                self.set(key, getattr(module, key), priority)

    def get_bool(self, name, default=False):
        got = self.get(name, default)
        try:
            return bool(int(got))
        except ValueError:
            if got in ("True", "true"):
                return True
            if got in ("False", "false"):
                return False
            raise ValueError("Supported values for boolean settings "
                             "are 0/1, True/False, '0'/'1', "
                             "'True'/'False' and 'true'/'false'")

    def init_settings(self, spider_ins):
        """
        设置spider的配置
        """

        for name in spider_attr_list:
            if getattr(spider_ins, name) is not None:
                self.set(name.upper(), getattr(spider_ins, name), "spider")

        for name in spider_attr_list:
            setattr(spider_ins, name, self.get(name.upper()))

    def print_log(self, engine_ins):
        setting_level = "project"
        if not self.project_setting:
            setting_level = "spider"

        common_list = ['name', 'worker_numbers', 'download_delay', 'run_forever',
                       'http_client_kwargs', 'log_level', 'log_write_file', 'serialization']

        body = f"\nsetting max level: {setting_level}"

        # blank
        blank = " " * 4
        for item in common_list:
            body += f"\n{blank}{item:28s}: {self.get(item.upper())}"

        queue_cls = self.get("QUEUE_CLS")
        queue_cls_str = const_map.get(queue_cls, queue_cls)
        body += f"\n{blank}{'queue_cls':28s}: {queue_cls_str}"
        if queue_cls == const.RedisQueue:
            body += f"\n{blank}{'clean_queue':28s}: {self.get('CLEAN_QUEUE')}"
            body += f"\n{blank}{'priority':28s}: {self.get('PRIORITY')}"
        elif queue_cls == const.RabbitMQQueue:
            body += f"\n{blank}{'clean_queue':28s}: {self.get('CLEAN_QUEUE')}"

        dupefilter_cls = self.get("DUPEFILTER_CLS")
        dupefilter_cls_str = const_map.get(dupefilter_cls, dupefilter_cls)
        body += f"\n{blank}{'dupefilter_cls':28s}: {dupefilter_cls_str}"
        if dupefilter_cls == const.RedisDupeFilter:
            body += f"\n{blank}{'clean_dupefilter':28s}: {self.get('CLEAN_DUPEFILTER')}"

        stats_cls = self.get("STATS_CLS")
        stats_cls_str = const_map.get(stats_cls, stats_cls)
        body += f"\n{blank}{'stats_cls':28s}: {stats_cls_str}"

        downloader_cls = self.get("DOWNLOADER_CLS")
        downloader_cls_str = const_map.get(downloader_cls, downloader_cls)
        body += f"\n{blank}{'downloader_cls':28s}: {downloader_cls_str}"

        downloader_middlewares = []
        spider_middlewares = []
        pipelines = []
        for item in self.get("DOWNLOADER_MIDDLEWARES"):
            downloader_middlewares.append(item.__module__)

        for item in self.get("SPIDER_MIDDLEWARES"):
            spider_middlewares.append(item.__module__)

        for item in self.get("PIPELINES"):
            pipelines.append(item.__module__)

        body += f"\n{blank}{'downloader_middlewares':28s}: {downloader_middlewares}"
        body += f"\n{blank}{'spider_middlewares':28s}: {spider_middlewares}"
        body += f"\n{blank}{'pipelines':28s}: {pipelines}"

        log_level = self.get("LOG_LEVEL").upper()
        logger.info(log_level)
        if log_level == "DEBUG":
            downloader_middlewares_base = []
            spider_middlewares_base = []
            pipelines_base = []

            for item in self.get("DOWNLOADER_MIDDLEWARES_BASE"):
                downloader_middlewares_base.append(item.__module__)

            for item in self.get("SPIDER_MIDDLEWARES_BASE"):
                spider_middlewares_base.append(item.__module__)

            for item in self.get("PIPELINES_BASE"):
                pipelines_base.append(item.__module__)

            body += f"\n{blank}{'downloader_middlewares_base':28s}: {downloader_middlewares_base}"
            body += f"\n{blank}{'spider_middlewares_base':28s}: {spider_middlewares_base}"
            body += f"\n{blank}{'pipelines_base':28s}: {pipelines_base}"

        logger.info(body)





"""
将list转成异步迭代器
"""


class AsyncIter:
    def __init__(self, items):
        self.items = items

    async def __aiter__(self):
        for item in self.items:
            yield item

    async def fun(self):
        for item in self.items:
            yield item




# Downloader:
import json

AiohttpDownloader = "hoopa.downloader.AiohttpDownloader"
HttpxDownloader = "hoopa.downloader.HttpxDownloader"
RequestsDownloader = "hoopa.downloader.RequestsDownloader"

# SchedulerQueue:
RedisQueue = "hoopa.queues.RedisQueue"
MemoryQueue = "hoopa.queues.MemoryQueue"
RabbitMQQueue = "hoopa.queues.RabbitMQQueue"

# Dupefilter:
RedisDupeFilter = "hoopa.dupefilters.RedisDupeFilter"
MemoryDupeFilter = "hoopa.dupefilters.MemoryDupeFilter"

# StatsCollector
MemoryStatsCollector = "hoopa.statscollectors.MemoryStatsCollector"  # memory
DummyStatsCollector = "hoopa.statscollectors.DummyStatsCollector"  # 假的，不进行统计
RedisStatsCollector = "hoopa.statscollectors.RedisStatsCollector"  # redis


const_map = {
    "hoopa.downloader.AiohttpDownloader": "AiohttpDownloader",
    "hoopa.downloader.HttpxDownloader": "HttpxDownloader",
    "hoopa.queues.RedisQueue": "RedisQueue",
    "hoopa.queues.MemoryQueue": "MemoryQueue",
    "hoopa.queues.RabbitMQQueue": "RabbitMQQueue",
    "hoopa.dupefilters.RedisDupeFilter": "RedisDupeFilter",
    "hoopa.dupefilters.MemoryDupeFilter": "MemoryDupeFilter",
    "hoopa.statscollectors.MemoryStatsCollector": "MemoryStatsCollector",
    "hoopa.statscollectors.DummyStatsCollector": "DummyStatsCollector",
    "hoopa.statscollectors.RedisStatsCollector": "RedisStatsCollector"
}


# encoding: utf-8
"""
exception
"""


class InvalidOutput(Exception):
    pass


class InvalidUrl(Exception):
    pass


class InvalidCallbackResult(Exception):
    pass


class InvalidCallback(Exception):
    pass


class InvalidLogLevelError(Exception):
    pass


class SpiderHookError(Exception):
    pass


class UsageError(Exception):
    """To indicate a command-line usage error"""

    def __init__(self, *a, **kw):
        self.print_help = kw.pop('print_help', True)
        super(UsageError, self).__init__(*a, **kw)


class Error:
    def __init__(self, exception=None, stack=None):
        self.exception = exception
        self.stack = stack

    @property
    def name(self):
        return str(self.exception)




# encoding: utf-8
"""
response对象
"""
import codecs
from dataclasses import dataclass

import cchardet
import ujson
from aiohttp import helpers
from parsel import Selector

from hoopa.exceptions import Error
from hoopa.utils.url import get_location_from_history


@dataclass
class Response:
    def __init__(
        self,
        url: str = "",
        *,
        encoding: str = "",
        cookies=None,
        history=None,
        headers=None,
        status: int = -1,
        body: bytes = b'',
        text: str = "",
        error: Error = None,
        ok: int = 1
    ):
        self._url = url
        self._encoding = encoding
        self._headers = headers
        self._cookies = cookies
        self._history = history
        self._status = status

        self._body = body
        self._text = text

        self._error: Error = error  # 错误

        self._ok: int = ok  # 请求状态： 1成功；0失败，会进行重试；-1失败，不进行失败，直接进入失败队列

    @property
    def url(self):
        return self._url

    @property
    def encoding(self):
        return self._encoding

    @property
    def content(self):
        return self._body

    @property
    def text(self, errors='ignore'):
        if self._text:
            return self._text

        if not self._body:
            return None

        if not self._encoding:
            self._encoding = self.get_encoding()

        return self._body.decode(self._encoding, errors=errors)

    @text.setter
    def text(self, value: str):
        self._text = value

    def json(self, *args, **kwargs):
        return ujson.loads(self.text, *args, **kwargs)

    @property
    def ok(self):
        return self._ok

    @ok.setter
    def ok(self, value: int):
        self._ok = value

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, value: str):
        self._error = value

    @property
    def headers(self):
        return self._headers

    @property
    def body(self):
        return self._body

    @property
    def cookies(self):
        return self._cookies

    @property
    def history(self):
        return self._history

    @property
    def status(self):
        return self._status

    @property
    def selector(self):
        return Selector(self.text)

    def xpath(self, xpath_str):
        return self.selector.xpath(xpath_str)

    def re(self, re_str):
        return self.selector.re(re_str)

    def css(self, re_str):
        return self.selector.css(re_str)

    def serialize(self):
        request_dict = {}
        for item in ["url", "status", "body", 'text', 'encoding', "ok"]:
            request_dict.setdefault(item, getattr(self, item))
        return request_dict

    def get_encoding(self) -> str:
        c_type = self._headers.get("Content-Type", "").lower()
        mimetype = helpers.parse_mimetype(c_type)

        encoding = mimetype.parameters.get("charset")
        if encoding:
            try:
                codecs.lookup(encoding)
            except LookupError:
                encoding = None
        if not encoding:
            if mimetype.type == "application" and (mimetype.subtype == "json" or mimetype.subtype == "rdap"):
                encoding = "utf-8"
            elif self._body is None:
                raise RuntimeError("Cannot guess the encoding of " "a not yet read body")
            else:
                encoding = cchardet.detect(self._body)["encoding"]
        if not encoding:
            encoding = "utf-8"

        return encoding

    @property
    def response_url(self):
        if self._history:
            last_res_url = get_location_from_history(self._history)
            return last_res_url
        else:
            return self.url

    def __repr__(self):
        return f"<Response [{self._status}]>"



import platform
import sys
from loguru import logger

from hoopa.exceptions import InvalidLogLevelError


class Logging:
    """
    日志模块, 使用loguru作为日志模块，弊端就是全局只能存在一个logger，如果想自定义，通过LOG_CONFIG设置
    """

    def __init__(self, setting=None):

        # 自定义logger的配置
        configure = setting.get("LOG_CONFIG", None)
        if configure:
            logger.configure(**configure)
            return

        log_write_file = setting.get_bool("LOG_WRITE_FILE")
        self.log_level = setting.get("LOG_LEVEL", "INFO").upper()

        if self.log_level not in ("TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"):
            raise InvalidLogLevelError(self.log_level)

        if self.log_level == "DEBUG":
            logger_format = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <7}</level> | <cyan>{name}" \
                            "</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"
        else:
            logger_format = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <7}</level> | <level>" \
                            "{message}</level>"

        if log_write_file:
            spider_name = setting.get("NAME")
            handler = {
                    "sink": "logs/%s.log" % spider_name,
                    "rotation": "00:00",
                    "level": self.log_level,
                    "enqueue": True,
                    "colorize": True,
                    "backtrace": False, "diagnose": False,
                    "format": logger_format,
                }
        else:
            handler = {
                "sink": sys.stderr,
                "level": self.log_level,
                "enqueue": True,
                "colorize": True,
                "format": logger_format,
                "backtrace": False, "diagnose": False,
            }

        config = {
            "handlers": [handler]
        }
        logger.configure(**config)

    def get_tb_limit(self):
        if self.log_level == "DEBUG":
            return None
        else:
            return -1


def ignore_windows_close_loop_error():
    """
    忽视windows下关闭事件循环的错误
    @return:
    """
    from functools import wraps
    from asyncio.proactor_events import _ProactorBasePipeTransport

    def silence_event_loop_closed(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except RuntimeError as e:
                if str(e) != 'Event loop is closed':
                    raise

        return wrapper

    if platform.system() == 'Windows':
        _ProactorBasePipeTransport.__del__ = silence_event_loop_closed(_ProactorBasePipeTransport.__del__)


ignore_windows_close_loop_error()


import os, sys
import re


sys.path.insert(0, re.sub(r"([\\/]items)|([\\/]spiders)", "", os.getcwd()))

__all__ = [
    "Spider",
    "Item",
    "Request",
    "Response",
    "helpers",
    "const",
    "Logging",
    "Setting",
    "AiohttpDownloader",
    "HttpxDownloader",
]

from hoopa.core.spider import Spider, RedisSpider
from hoopa.item import Item
from hoopa.request import Request
from hoopa.response import Response
from hoopa.utils import helpers
from hoopa.settings import const
from hoopa.utils.log import Logging
from hoopa.utils.project import Setting
from hoopa.downloader import AiohttpDownloader, HttpxDownloader

__version__ = "0.1.8"



from loguru import logger

from hoopa.request import Request


class HandleParseErrorMiddleware:
    async def process_exception(self, request: Request, response, error, spider_ins):
        response.ok = 0
        response.error = error
        logger.error(f"{request.callback} 解析出现错误:\nrequest: {request.serialize()}\n{response}\n {error.stack}")
        return True





# encoding: utf-8
import asyncio
import hashlib
import os
import random
import uuid
from asyncio import iscoroutinefunction
from importlib import import_module

import arrow

from hoopa.utils.concurrency import run_function_no_concurrency


def get_md5(data):
    return hashlib.md5(data.encode(encoding='UTF-8')).hexdigest()


def get_timestamp():
    return arrow.now().float_timestamp


def get_datetime(fmt="YYYY-MM-DD HH:MM:SS"):
    return arrow.now().format(fmt)


def split_list(seq, step):
    return [seq[i:i+step] for i in range(0, len(seq), step)]


def load_object(path):
    try:
        dot = path.rindex('.')
    except ValueError:
        raise ValueError("Error loading object '%s': not a full path" % path)

    module, name = path[:dot], path[dot + 1:]
    mod = import_module(module)

    try:
        obj = getattr(mod, name)
    except AttributeError:
        raise NameError("Module '%s' doesn't define any object named '%s'" % (module, name))

    return obj


async def create_instance(objcls, *args, **kwargs):
    if hasattr(objcls, 'create'):
        instance = await run_function_no_concurrency(objcls.create, *args, **kwargs)
    else:
        instance = objcls()
    return instance


async def create_instance_and_init(objcls, engine, *args, **kwargs):
    instance = await create_instance(objcls, engine, *args, **kwargs)
    if hasattr(instance, 'init'):
        if instance.init.__code__.co_argcount == 1:
            await run_function_no_concurrency(instance.init)
        else:
            await run_function_no_concurrency(instance.init, engine.spider)

    return instance


def request_fingerprint(request):
    fp = hashlib.md5()
    fp.update(to_bytes(request.method))
    fp.update(to_bytes(request.request_url))
    fp.update(to_bytes(str(request.data)) or b'')
    fp.update(to_bytes(str(request.json)) or b'')

    return fp.hexdigest()


async def spider_sleep(download_delay):
    """
    执行休眠
    @return:
    """
    # 休眠间隔, 加入随机值选项
    if isinstance(download_delay, (list, tuple)):
        download_delay = random.uniform(download_delay[0], download_delay[1])
    else:
        download_delay = download_delay

    await asyncio.sleep(max(download_delay, 0.01))


def to_str(bytes_or_str, encoding="utf-8", errors='strict'):
    if isinstance(bytes_or_str, bytes):
        value = bytes_or_str.decode(encoding, errors)
    else:
        value = bytes_or_str
    return value


def to_bytes(bytes_or_str, encoding="utf-8", errors='strict'):
    if isinstance(bytes_or_str, str):
        value = bytes_or_str.encode(encoding, errors)
    else:
        value = bytes_or_str
    return value


def get_uuid():
    return str(uuid.uuid4())


def get_continue_list(p_list):
    # 去重排序
    seq = sorted(list(set(p_list)))
    result_seq = []
    tmp = None
    tmp_list = []
    for i in seq:
        if tmp is None:
            tmp_list.append(i)
        elif i - tmp == 1:
            tmp_list.append(i)
        else:
            result_seq.append(tmp_list)
            tmp_list = [i]

        tmp = i
    if tmp_list:
        result_seq.append(tmp_list)
    return result_seq


def get_priority_list(p_list):
    continue_list = get_continue_list(p_list)
    priority_list = []

    for item in continue_list:
        priority_list.append((item[0], item[-1]))

    return priority_list


def get_mac_address():
    node = uuid.getnode()
    mac = uuid.UUID(int=node).hex[-12:]
    return mac


def get_pid():
    return os.getpid()


def get_mac_pid():
    mac = get_mac_address()
    pid = get_pid()
    return f"{mac}#{pid}"



#!/usr/bin/env python
import os
import re
import sys

from setuptools import find_packages, setup


_version_re = re.compile(r"__version__\s+=\s+(.*)")

PY_VER = sys.version_info

if PY_VER < (3, 7):
    raise RuntimeError("hoopa doesn't support Python version prior 3.7")


def read_version():
    regexp = re.compile(r'^__version__\W*=\W*"([\d.abrc]+)"')
    init_py = os.path.join(os.path.dirname(__file__), "hoopa", "__init__.py")
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)


def read(file_name):
    with open(
        os.path.join(os.path.dirname(__file__), file_name), mode="r", encoding="utf-8"
    ) as f:
        return f.read()


requires = [
    "aiohttp>=3.9.5",
    "httpx[http2]>=0.17.0",
    "aioredis>=1.3.1",
    "requests",
    "parsel",
    "aiodns",
    "cchardet",
    "arrow",
    "w3lib",
    "loguru",
    "ujson",
    "lxml",
]


setup(
    name="hoopa",
    version=read_version(),
    author="fishtn",
    description="Asynchronous crawler micro-framework based on python.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author_email="",
    python_requires=">=3.9",
    install_requires=requires,
    url="https://github.com/fishtn/hoopa",
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: BSD",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    extras_require={"uvloop": ["uvloop"]},
    entry_points={"console_scripts": ["hoopa = hoopa.commands.cmdline:execute"]},
    include_package_data=True
)



#!/usr/bin/env python
from collections import deque, defaultdict

from hoopa.utils.concurrency import run_function
from hoopa.utils.helpers import create_instance_and_init


class MiddlewareManager:
    """中间件管理类"""

    def __init__(self,  middlewares, engine=None):
        self.engine = engine

        # 存放中间件队列的字典
        self.methods = defaultdict(deque)
        # 存放中间件名称队列的字典
        self.names = defaultdict(deque)

        for mw in middlewares:
            self._add_middleware(mw)

    @classmethod
    async def create(cls, engine):
        mw_list = cls._get_mw_list_from_engine(engine)

        middlewares = []
        for mw_cls in mw_list:
            mw = await create_instance_and_init(mw_cls, engine)
            middlewares.append(mw)

        return cls(middlewares, engine)

    @classmethod
    def _get_mw_list_from_engine(cls, setting):
        raise NotImplementedError

    def _add_middleware(self, mw):
        """
        add init and close
        @param mw: middleware
        @return:
        """
        if hasattr(mw, "init") and callable(getattr(mw, "init")):
            self.methods["init"].appendleft(mw.init)

        if hasattr(mw, "close") and callable(getattr(mw, "close")):
            self.methods["close"].appendleft(mw.close)

    async def _process_chain(self, method_name, obj, *args):
        for method in self.methods[method_name]:
            await run_function(method, obj, *args)

    async def init(self):
        await self._process_chain("init", self.engine.spider)

    async def close(self):
        await self._process_chain("close", self.engine.spider)



from urllib.parse import urlparse

import aioredis


def get_redis_uri_from_dict(**kwargs):
    """
    将redis连接字典转成uri
    @param kwargs: 参数字典，仅含host, port, db, password, encoding
    @return:
    """
    host = kwargs.pop('host', None)
    port = kwargs.pop('port', None)
    db = kwargs.pop('db', None)
    password = kwargs.pop('password', None)
    encoding = kwargs.pop('encoding', "utf-8")

    if password:
        uri = f"redis://:{password}@{host}:{port}/{db}?encoding={encoding}"
    else:
        uri = f"redis://{host}:{port}/{db}?encoding={encoding}"

    return uri


async def get_aio_redis(redis_setting):
    """
    创建aioredis连接池
    @param redis_setting: redis配置dict或者uri
    @return:
    """
    uri = redis_setting
    if isinstance(redis_setting, dict):
        uri = get_redis_uri_from_dict(**redis_setting)
    return await aioredis.create_redis_pool(uri)


def get_host(redis_setting):
    """
    从redis配置字典和uri获取host
    @param redis_setting:
    @return:
    """
    if isinstance(redis_setting, dict):
        return redis_setting["host"]
    else:
        return urlparse(redis_setting).hostname



# -*- coding: utf-8 -*-
"""
装饰器
"""

import asyncio
from functools import wraps

from loguru import logger

from .helpers import spider_sleep
from ..request import Request


def handle_download_callback_retry():
    def retry(func):
        @wraps(func)
        async def wrapper(self, request: Request):
            _retry = 0
            try:
                while True:
                    response = await func(self, request)

                    # 成功返回response
                    if response.ok == 1:
                        return response

                    # 重试次数大于等于最大重试次数
                    if request.retries >= request.retry_times:
                        logger.error(f"{request} too many error, try {request.retries} times")
                        response.ok = -1

                    # 当ok == -1的时候，直接返回
                    if response.ok == -1:
                        return response

                    # 重试次数加1
                    request.retries += 1

                    if request.retries == 1:
                        self.retrying += 1
                        _retry = 1

                    # 统计重试次数
                    stats_name = response.error.name if response.error else response.status
                    await self.stats.inc_value(f"requests/retries/{stats_name}", 1)

                    # 休眠
                    # logger.error(f"request.retry_delay {request.retry_delay}")
                    await asyncio.sleep(request.retry_delay)
            finally:
                self.retrying -= _retry

        return wrapper

    return retry


def timeout_it(timeout=600):
    def __timeout_it(func):
        @wraps(func)
        async def wrapper(self, request: Request):
            try:
                await asyncio.wait_for(func(self, request), timeout)
            except asyncio.TimeoutError:
                logger.error(f"task timeout: {request}")
            except Exception as e:
                logger.error(f"error {request}\n: {e}")

        return wrapper
    return __timeout_it




class StatsMiddleware:

    def __init__(self, stats):
        self.stats = stats

    @classmethod
    async def create(cls, engine):
        return cls(engine.stats)

    async def process_request(self, request, spider_ins):
        await self.stats.inc_value('downloader/request_method_count/%s' % request.method)

    async def process_response(self, request, response, spider_ins):
        await self.stats.inc_value('downloader/response_status_count/%s' % response.status)







# encoding: utf-8
"""
爬虫队列
"""
import asyncio
import importlib
import time
import traceback
import typing
from asyncio import PriorityQueue

import ujson
from loguru import logger

from hoopa.request import Request
from hoopa.response import Response
from hoopa.utils.connection import get_aio_redis
from hoopa.utils.helpers import get_timestamp, get_priority_list, get_mac_pid


class BaseQueue:
    async def get(self, priority):
        """
        从队列中获取一个request
        @param priority: 权重，取出对应权重的request
        """
        pass

    async def add(self, requests: typing.Union[Request, typing.List[Request]]):
        """
        向队列添加多个request
        @param requests:
        """
        pass

    async def set_result(self, request: Request, response: Response, task_request: Request):
        """
        保存结果
        @param request:
        @param response:
        @param task_request:
        """
        pass

    async def clean_queue(self):
        """
        清空队列
        """
        pass

    async def check_status(self, spider_ins, run_forever):
        """
        检查状态，主要是判断爬虫是否爬取完毕，其他的具体根据队列自身实现需要的功能
        @return:
        """
        pass

    async def close(self):
        pass


class MemoryQueue(BaseQueue):
    def __init__(self, waiting, serialization_module, engine):
        # 下载队列，一个优先级队列
        self.waiting = waiting
        # 进行中的队列，key为Request，value取出的时间戳
        self.pending = {}
        # 失败次数记录，key为Request，value失败次数
        self.failure = {}
        self.serialization_module = serialization_module
        self.engine = engine

    @classmethod
    async def create(cls, engine):
        waiting = PriorityQueue()
        serialization_module = importlib.import_module(engine.setting["SERIALIZATION"])
        return cls(waiting, serialization_module, engine)

    async def clean_scheduler(self, waiting=True, pending=True, failure=True, data=True):
        """
        清空队列
        """
        pass

    async def get(self, priority):
        """
        从队列中获取一个request
        """
        if not self.waiting.empty():
            result = await self.waiting.get()
            self.pending[result[1]] = get_timestamp()
            return Request.unserialize(result[1], self.serialization_module)
        return None

    async def add(self, requests: typing.Union[Request, typing.List[Request]]):
        """
        向队列添加多个request
        @param requests:
        """
        if isinstance(requests, Request):
            requests = [requests]

        count = 0
        # 判断是否在pending中，如果在，是否过了最大时间
        for request in requests:
            str_request = request.serialize(self.serialization_module)
            pended_time = self.pending.get(str_request, 0)
            if time.time() - pended_time < self.engine.setting["PENDING_THRESHOLD"]:
                continue

            count += 1
            self.waiting.put_nowait((-request.priority, str_request))
            if pended_time:
                self.pending.pop(str_request)
        return count

    async def set_result(self, request: Request, response: Response, task_request: Request):
        """
        保存结果
        @param request:
        @param response:
        @param task_request:
        """
        # 如果失败，且失败次数未达到，返回waiting
        str_request = request.serialize(self.serialization_module)

        # 如果在进行队列中，删除
        if str_request in self.pending:
            self.pending.pop(str_request)

        # 如果成功
        if response.ok == 1:
            return True

        if response.ok == -1:
            self.failure[str_request] = response.status
            return False

        if str_request in self.failure:
            self.failure[str_request] += 1
            await self.add(request)
        else:
            self.failure[str_request] = 1
            await self.add(request)

    async def check_status(self, spider_ins, run_forever=False):
        if not self.pending and self.waiting.empty():
            spider_ins.run = False


class RedisQueue(BaseQueue):
    """
    Redis队列
    """
    def __init__(self, spider_name, serialization_module, engine):
        self._spider_name = spider_name
        self.serialization_module = serialization_module
        self.engine = engine
        
        self._failure_key = f"{spider_name}:failure"
        self._pending_key = f"{spider_name}:pending"
        self._waiting_key = f"{spider_name}:waiting"
        self._client_key = f"{spider_name}:client"
        
        self.pool = None
        self._last_check_pending_task_time = 0
        
        # 统计信息
        self.task_count = 0
        self.task_success = 0
        self.task_failure = 0

    @classmethod
    async def create(cls, engine):
        serialization_module = importlib.import_module(engine.setting["SERIALIZATION"])
        spider_name = engine.setting["NAME"]
        return cls(spider_name, serialization_module, engine)

    async def init(self):
        """
        初始化db
        """
        self.pool = await get_aio_redis(self.engine.setting["REDIS_SETTING"])
        loop = asyncio.get_running_loop()
        asyncio.run_coroutine_threadsafe(self.set_heart_beat(), loop=loop)

    async def set_heart_beat(self):
        start_time = int(time.time())
        last_time_requests_count = 0
        while True:

            now_time = int(time.time())
            run_time = int(now_time - start_time)
            if run_time:
                requests_count = self.engine.requests_count
                requests_per_second_all = round(requests_count / run_time, 1)
                requests_per_second_10s = round((requests_count - last_time_requests_count) / 10, 1)
                last_time_requests_count = requests_count
                with await self.pool as conn:
                    data = {
                        "总数": self.task_count,
                        "成功": self.task_success,
                        "失败": self.task_failure,
                        "请求": self.engine.requests_count,
                        "每秒请求(全部)": requests_per_second_all,
                        "每秒请求(10s)": requests_per_second_10s,
                        "上报时间": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(now_time)),
                    }
                    dumps_data = ujson.dumps(data, ensure_ascii=False)
                    logger.info(f"统计: {dumps_data}")
                    await conn.hset(self._client_key, get_mac_pid(), dumps_data)

            await asyncio.sleep(10)

    async def clean_queue(self):
        """
        清空队列
        """
        with await self.pool as conn:
            # 要避免一次性删除过大的key，导致redis阻塞
            await conn.delete(self._failure_key)
            await conn.delete(self._pending_key)
            await conn.delete(self._waiting_key)

    async def get(self, priority: typing.Union[int, list]):
        """
        从redis中获取request
        @param priority: 为None的时候，获取所有权重，否则获取指定的权重，可以是int，也可以是int列表
        @return: request
        """
        priority_list = []
        if priority is None:
            priority_list.append(("-inf", "+inf"))
        elif isinstance(priority, int):
            priority_list.append((priority, priority))
        else:
            priority_list = get_priority_list(priority)

        try:
            lua = """
                redis.replicate_commands()
                local waiting_key = KEYS[1]
                local pending_key = KEYS[2]
                local min = KEYS[3]
                local max = KEYS[4]

                -- 取值
                local result = redis.call('zrevrangebyscore', waiting_key, max, min, 'LIMIT', 0, 1)

                if result and table.getn(result) > 0 then
                    redis.call('zrem', waiting_key, result[1])
                    redis.call('hset', pending_key, result[1], redis.call('TIME')[1])
                    return result[1]
                end
                return nil
            """
            with await self.pool as conn:
                for p_item in priority_list:
                    _min, _max  = p_item
                    eval_result = await conn.eval(lua, keys=[self._waiting_key, self._pending_key, _min, _max], args=[])
                    if eval_result:
                        self.task_count += 1
                        return Request.unserialize(eval_result, self.serialization_module)
        except Exception as e:
            logger.error(f"get request error \n{traceback.format_exc()}")

        return None

    async def add(self, requests):
        """
        向队列添加request
        @param requests: request列表
        @return:
        """
        if not isinstance(requests, list):
            requests = [requests]

        str_requests = [_.serialize(self.serialization_module) for _ in requests]
        priority_list = [_.priority for _ in requests]
        lua = """
            redis.replicate_commands()
            local priority_list = KEYS
            local requests = ARGV

            local spider = table.remove(priority_list, 1)

            local now = redis.call('TIME')[1]
            local waiting_key = spider..':waiting'
            local pending_key = spider..':pending'

            local add_counts = 0
            for i, v in ipairs(requests) do
                -- 判断在pending中的时间
                local score = redis.call('hget', pending_key, v)
                if (score and tonumber(now) - tonumber(score) >= 30) or (not score) then
                    local result = redis.call('zadd', waiting_key, priority_list[i], v)
                    add_counts = add_counts + result
                    redis.call('hdel', pending_key, v)
                end
            end

            return add_counts
        """
        with await self.pool as conn:
            add_counts = await conn.eval(lua, keys=[self._spider_name] + priority_list, args=str_requests)
        return add_counts

    async def set_result(self, request: Request, response: Response, task_request: Request):
        """
        保存结果，设置状态（成功或失败）
        @param request:
        @param response:
        @param task_request:
        @return:
        """

        request_ser = request.serialize(self.serialization_module)
        with await self.pool as conn:
            if response.ok == 1:
                # 成功，删除pending队列
                await conn.hdel(self._pending_key, request_ser)
                self.task_success += 1
            else:
                # 失败, 从等待队列中删除，并放到失败队列
                pipe = conn.pipeline()
                pipe.hdel(self._pending_key, request_ser)
                pipe.hset(self._failure_key, request_ser, response.status)
                await pipe.execute()
                self.task_failure += 1

    async def check_status(self, spider_ins, run_forever=False):
        with await self.pool as conn:
            pending_len = await conn.hlen(self._pending_key)
            waiting_len = await conn.zcard(self._waiting_key)
            if not pending_len and not waiting_len:
                spider_ins.run = False

        await self.check_pending_task()

    async def check_pending_task(self):
        # 判断是否有超时的链接
        now_time = time.time()
        if now_time - self._last_check_pending_task_time > 10:
            self._last_check_pending_task_time = now_time
            now_time = time.time()

            with await self.pool as conn:
                pending_list = await conn.hgetall(self._pending_key)

                to_waiting_list = []
                del_pending_list = []
                for k, v in pending_list.items():
                    if now_time - int(v) > self.engine.setting["PENDING_THRESHOLD"]:
                        request = Request.unserialize(k, self.serialization_module)
                        to_waiting_list.extend([request.priority, k])
                        del_pending_list.append(k)

                if to_waiting_list:
                    pipe = conn.pipeline()
                    pipe.zadd(self._waiting_key, *to_waiting_list)
                    pipe.hdel(self._pending_key, *del_pending_list)
                    result = await pipe.execute()

                    logger.info(f"pendings: {len(pending_list)}, del_pending: {result[1]}, add_waitings: {result[0]}")

    async def failure_to_waiting(self, spider_ins):

        with await self.pool as conn:
            failure_list: dict = await conn.hgetall(self._failure_key)

        if failure_list:
            zadd_list = []
            hdel_list = []
            for key, value in failure_list.items():
                request = Request.unserialize(key)
                zadd_list.extend([request.priority, key])
                hdel_list.append(key)

            if zadd_list:
                with await self.pool as conn:
                    try:
                        pipe = conn.pipeline()
                        pipe.zadd(self._waiting_key, *zadd_list)
                        pipe.hdel(self._failure_key, *hdel_list)
                        await pipe.execute()
                    except:
                        logger.debug(traceback.format_exc())
                        logger.error("failure_to_waiting error")

                    logger.info(f"failure_to_waiting, result: {len(hdel_list)}")

    async def close(self):
        self.pool.close()



# encoding: utf-8
"""
下载器
"""

import aiohttp
import httpx
import requests
from aiohttp import TCPConnector

from hoopa.request import Request
from hoopa.response import Response
from hoopa.utils.concurrency import run_function


class Downloader:
    """
    下载器基础类，子类需要实现：init, close, get_session, fetch
    同步调用：fetch
    """
    session = None

    async def close(self):
        pass

    async def get_session(self, request):
        """
        1. 优先使用request参数的session
        2. client_kwargs参数不为空，新建一个session
        3. 全局session不为空，使用全局session
        4. 新建一个session，提供给单次下载
        """
        pass

    def fetch(self, request: Request) -> Response:
        """
        同步请求
        @param request:
        @return: Response
        """
        pass


class AiohttpDownloader(Downloader):
    """
    Aiohttp下载器
    """
    def __init__(self, http_client_kwargs, engine):
        self.http_client_kwargs = http_client_kwargs
        self.engine = engine

        self.tc = None

    @classmethod
    async def create(cls, engine):
        http_client_kwargs = engine.setting["HTTP_CLIENT_KWARGS"]
        return cls(http_client_kwargs, engine)

    async def init(self):
        if self.http_client_kwargs:
            self.session = aiohttp.ClientSession(**self.http_client_kwargs)
        else:
            jar = aiohttp.DummyCookieJar()
            self.tc = TCPConnector(limit=100, force_close=True, enable_cleanup_closed=True, verify_ssl=False)
            self.session = aiohttp.ClientSession(connector=self.tc, cookie_jar=jar)

    async def close(self):
        if self.tc:
            await self.tc.close()
        if self.session:
            await self.session.close()

    async def get_session(self, request):
        """
        1. 优先使用request参数的session
        2. client_kwargs参数不为空，新建一个session
        3. 使用全局session
        """
        is_close = False
        if request.session:
            session = request.session
        elif request.client_kwargs:
            # 单个请求创建的会话需要使用后关闭
            is_close = True
            session = aiohttp.ClientSession(**request.client_kwargs)
        else:
            session = self.session
        return session, is_close

    async def fetch(self, request: Request) -> Response:
        session, is_close = await self.get_session(request)
        _kwargs = request.replace_to_kwargs
        try:
            async with session.request(**_kwargs) as resp:
                response = Response(
                    url=str(resp.url),
                    body=await resp.read(),
                    status=resp.status,
                    cookies=resp.cookies,
                    headers=resp.headers,
                    history=resp.history,
                )
                return response
        finally:
            if is_close:
                await session.close()


class HttpxDownloader(Downloader):
    """
    Httpx下载器
    """

    def __init__(self, http_client_kwargs, engine):
        self.http_client_kwargs = http_client_kwargs
        self.engine = engine

    @classmethod
    async def create(cls, engine):
        http_client_kwargs = engine.setting["HTTP_CLIENT_KWARGS"]
        return cls(http_client_kwargs, engine)

    async def init(self):
        if self.http_client_kwargs:
            self.session = httpx.AsyncClient(**self.http_client_kwargs)
        else:
            self.session = httpx.AsyncClient(http2=True, verify=False)

    async def close(self):
        await self.session.aclose()

    async def fetch(self, request: Request) -> Response:
        session, is_close = await self.get_session(request)
        _kwargs = request.replace_to_kwargs

        try:
            resp = await self.session.request(**_kwargs)
            response = Response(
                url=str(resp.url),
                body=resp.content,
                status=resp.status_code,
                cookies=resp.cookies,
                headers=resp.headers,
                history=resp.history,
            )
            return response
        finally:
            if is_close:
                await session.aclose()

    async def get_session(self, request):
        is_close = False
        if request.session:
            session = request.session
        elif request.client_kwargs:
            # 单个请求创建的会话需要使用后关闭
            is_close = True
            session = httpx.AsyncClient(**request.client_kwargs)
        else:
            session = self.session

        return session, is_close


class RequestsDownloader(Downloader):
    """
    Requests下载器
    """

    def __init__(self, http_client_kwargs, engine):
        self.http_client_kwargs = http_client_kwargs
        self.engine = engine

    @classmethod
    async def create(cls, engine):
        http_client_kwargs = engine.setting["HTTP_CLIENT_KWARGS"]
        return cls(http_client_kwargs, engine)

    def init(self):
        session = requests.Session()
        if self.http_client_kwargs:
            self.session = self.set_session(session, **self.http_client_kwargs)
        else:
            self.session = self.set_session(session)

    def close(self):
        self.session.close()

    async def fetch(self, request: Request) -> Response:
        return await run_function(self.sync_fetch, request)

    def sync_fetch(self, request: Request) -> Response:
        session, is_close = self.get_session(request)
        _kwargs = request.replace_to_kwargs

        try:
            resp = self.session.request(**_kwargs)
            response = Response(
                url=str(resp.url),
                body=resp.content,
                status=resp.status_code,
                cookies=resp.cookies,
                headers=resp.headers,
                history=resp.history
            )
            return response
        finally:
            if is_close:
                session.close()

    def get_session(self, request):
        is_close = False
        if request.session:
            session = request.session
        elif request.client_kwargs:
            # 单个请求创建的会话需要使用后关闭
            session = requests.Session()
            session = self.set_session(session, **request.client_kwargs)
        else:
            session = requests.Session()
        return session, is_close

    @staticmethod
    def set_session(session, **kwargs):
        for key, value in kwargs.items():
            if hasattr(session, key):
                setattr(session, key, value)
        return session
